# -*- coding: utf-8 -*-

import logging
import os
from os.path import join
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	import io
	from collections.abc import Callable

	from .glossary_types import GlossaryType


stdCompressions = ("gz", "bz2", "lzma")

log = logging.getLogger("pyglossary")

__all__ = [
	"compress",
	"compressionOpen",
	"compressionOpenFunc",
	"stdCompressions",
	"uncompress",
	"zipFileOrDir",
]


def compressionOpenFunc(c: str) -> "Callable | None":
	if not c:
		return open
	if c == "gz":
		import gzip

		return gzip.open
	if c == "bz2":
		import bz2

		return bz2.open
	if c == "lzma":
		import lzma

		return lzma.open
	if c == "dz":
		import gzip

		return gzip.open
	return None


def compressionOpen(
	filename: str,
	dz: bool = False,
	**kwargs,  # noqa: ANN003
) -> "io.IOBase":
	from os.path import splitext

	filenameNoExt, ext = splitext(filename)
	ext = ext.lower().lstrip(".")
	try:
		int(ext)
	except ValueError:
		pass
	else:
		_, ext = splitext(filenameNoExt)
		ext = ext.lower().lstrip(".")
	if ext in stdCompressions or (dz and ext == "dz"):
		openFunc = compressionOpenFunc(ext)
		if not openFunc:
			raise RuntimeError(f"no compression found for {ext=}")
		_file = openFunc(filename, **kwargs)
		_file.compression = ext
		return _file
	return open(filename, **kwargs)  # noqa: SIM115


def zipFileOrDir(filename: str) -> None:
	import shutil
	import zipfile
	from os.path import (
		isdir,
		isfile,
		split,
	)

	from .os_utils import indir

	def _zipFileAdd(zf: "zipfile.ZipFile", filename: str) -> None:
		if isfile(filename):
			zf.write(filename)
			return
		if not isdir(filename):
			raise OSError(f"Not a file or directory: {filename}")
		for subFname in os.listdir(filename):
			_zipFileAdd(zf, join(filename, subFname))

	with zipfile.ZipFile(f"{filename}.zip", mode="w") as zf:
		if isdir(filename):
			dirn, name = split(filename)
			with indir(filename):
				for subFname in os.listdir(filename):
					_zipFileAdd(zf, subFname)

			shutil.rmtree(filename)
			return

		dirn, name = split(filename)
		files = [name]

		if isdir(f"{filename}_res"):
			files.append(f"{name}_res")

		with indir(dirn):
			for fname in files:
				_zipFileAdd(zf, fname)


def compress(_glos: "GlossaryType", filename: str, compression: str) -> str:
	"""
	Filename is the existing file path.

	supported compressions: "gz", "bz2", "lzma", "zip".
	"""
	import shutil
	from os.path import isfile

	log.info(f"Compressing {filename!r} with {compression!r}")

	compFilename = f"{filename}.{compression}"
	if compression in stdCompressions:
		openFunc = compressionOpenFunc(compression)
		if not openFunc:
			raise RuntimeError(f"invalid {compression=}")
		with openFunc(compFilename, mode="wb") as dest:
			with open(filename, mode="rb") as source:
				shutil.copyfileobj(source, dest)
		return compFilename

	if compression == "zip":
		try:
			os.remove(compFilename)
		except OSError:
			pass
		try:
			zipFileOrDir(filename)
		except Exception as e:
			log.error(
				f'{e}\nFailed to compress file "{filename}"',
			)
	else:
		raise ValueError(f"unexpected {compression=}")

	if isfile(compFilename):
		return compFilename

	return filename


def uncompress(srcFilename: str, dstFilename: str, compression: str) -> None:
	"""
	Filename is the existing file path.

	supported compressions: "gz", "bz2", "lzma".
	"""
	import shutil

	log.info(f"Uncompressing {srcFilename!r} to {dstFilename!r}")

	if compression in stdCompressions:
		openFunc = compressionOpenFunc(compression)
		if not openFunc:
			raise RuntimeError(f"invalid {compression=}")
		with openFunc(srcFilename, mode="rb") as source:
			with open(dstFilename, mode="wb") as dest:
				shutil.copyfileobj(source, dest)
		return

	# TODO: if compression == "zip":
	raise ValueError(f"unsupported compression {compression!r}")



# -*- coding: utf-8 -*-

from collections.abc import Generator

from pyglossary.compression import stdCompressions
from pyglossary.core import log
from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.option import (
	BoolOption,
	EncodingOption,
	FileSizeOption,
	Option,
)
from pyglossary.text_reader import TextGlossaryReader
from pyglossary.text_utils import (
	splitByBarUnescapeNTB,
	unescapeNTB,
)

__all__ = ["Reader"]

enable = True
lname = "tabfile"
format = "Tabfile"
description = "Tabfile (.txt, .dic)"
extensions = (".txt", ".tab", ".tsv")
extensionCreate = ".txt"
singleFile = True
kind = "text"
wiki = "https://en.wikipedia.org/wiki/Tab-separated_values"
website = None
optionsProp: "dict[str, Option]" = {
	"encoding": EncodingOption(),
	"enable_info": BoolOption(
		comment="Enable glossary info / metedata",
	),
	"resources": BoolOption(
		comment="Enable resources / data files",
	),
	"file_size_approx": FileSizeOption(
		comment="Split up by given approximate file size\nexamples: 100m, 1g",
	),
	"word_title": BoolOption(
		comment="Add headwords title to beginning of definition",
	),
}


class Reader(TextGlossaryReader):
	@classmethod
	def isInfoWord(cls, word: str) -> bool:
		return word.startswith("#")

	@classmethod
	def fixInfoWord(cls, word: str) -> str:
		return word.lstrip("#")

	def nextBlock(self) -> "tuple[str | list[str], str, None] | None":
		if not self._file:
			raise StopIteration
		line = self.readline()
		if not line:
			raise StopIteration
		line = line.rstrip("\n")
		if not line:
			return None
		###
		word: "str | list[str]"
		word, tab, defi = line.partition("\t")
		if not tab:
			log.warning(
				f"Warning: line starting with {line[:10]!r} has no tab!",
			)
			return None
		###
		if self._glos.alts:
			word = splitByBarUnescapeNTB(word)
			if len(word) == 1:
				word = word[0]
		else:
			word = unescapeNTB(word, bar=False)
		###
		defi = unescapeNTB(defi)
		###
		return word, defi, None


class Writer:
	_encoding: str = "utf-8"
	_enable_info: bool = True
	_resources: bool = True
	_file_size_approx: int = 0
	_word_title: bool = False

	compressions = stdCompressions

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = ""

	def open(
		self,
		filename: str,
	) -> None:
		self._filename = filename

	def finish(self) -> None:
		pass

	def write(self) -> "Generator[None, EntryType, None]":
		from pyglossary.text_utils import escapeNTB, joinByBar
		from pyglossary.text_writer import TextGlossaryWriter

		writer = TextGlossaryWriter(
			self._glos,
			entryFmt="{word}\t{defi}\n",
			writeInfo=self._enable_info,
			outInfoKeysAliasDict=None,
		)
		writer.setAttrs(
			encoding=self._encoding,
			wordListEncodeFunc=joinByBar,
			wordEscapeFunc=escapeNTB,
			defiEscapeFunc=escapeNTB,
			ext=".txt",
			resources=self._resources,
			word_title=self._word_title,
			file_size_approx=self._file_size_approx,
		)
		writer.open(self._filename)
		yield from writer.write()
		writer.finish()



# -*- coding: utf-8 -*-

# Copyright © 2023 soshial <soshial@gmail.com> (soshial)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

from typing import NamedTuple

__all__ = ["ArticleAddress"]


class ArticleAddress(NamedTuple):
	sectionOffset: int
	chunkOffset: int

	def __str__(self) -> str:
		return f"Addr[{hex(self.sectionOffset)}, {hex(self.chunkOffset)}]"



import argparse
import logging
import os
from typing import Any

from pyglossary.ui.base import UIBase
from pyglossary.ui.option_ui import registerConfigOption


def defineFlags(parser: argparse.ArgumentParser, config: dict[str, Any]):
	defaultHasColor = config.get(
		"color.enable.cmd.windows" if os.sep == "\\" else "color.enable.cmd.unix",
		True,
	)
	parser.add_argument(
		"-v",
		"--verbosity",
		action="store",
		dest="verbosity",
		type=int,
		choices=(0, 1, 2, 3, 4, 5),
		required=False,
		default=int(os.getenv("VERBOSITY", "3")),
	)

	parser.add_argument(
		"--version",
		action="store_true",
	)
	parser.add_argument(
		"-h",
		"--help",
		dest="help",
		action="store_true",
	)
	parser.add_argument(
		"-u",
		"--ui",
		dest="ui_type",
		default="auto",
		choices=(
			"cmd",
			"gtk",
			"gtk4",
			"tk",
			# "qt",
			"auto",
			"none",
		),
	)
	parser.add_argument(
		"--cmd",
		dest="ui_type",
		action="store_const",
		const="cmd",
		default=None,
		help="use command-line user interface",
	)
	parser.add_argument(
		"--gtk",
		dest="ui_type",
		action="store_const",
		const="gtk",
		default=None,
		help="use Gtk-based user interface",
	)
	parser.add_argument(
		"--gtk4",
		dest="ui_type",
		action="store_const",
		const="gtk4",
		default=None,
		help="use Gtk4-based user interface",
	)
	parser.add_argument(
		"--tk",
		dest="ui_type",
		action="store_const",
		const="tk",
		default=None,
		help="use Tkinter-based user interface",
	)
	parser.add_argument(
		"--interactive",
		"--inter",
		dest="interactive",
		action="store_true",
		default=None,
		help="switch to interactive command line interface",
	)
	parser.add_argument(
		"--no-interactive",
		"--no-inter",
		dest="no_interactive",
		action="store_true",
		default=None,
		help=(
			"do not automatically switch to interactive command line"
			" interface, for scripts"
		),
	)

	parser.add_argument(
		"-r",
		"--read-options",
		dest="readOptions",
		default="",
	)
	parser.add_argument(
		"-w",
		"--write-options",
		dest="writeOptions",
		default="",
	)

	parser.add_argument(
		"--json-read-options",
		dest="jsonReadOptions",
		default=None,
	)
	parser.add_argument(
		"--json-write-options",
		dest="jsonWriteOptions",
		default=None,
	)

	parser.add_argument(
		"--read-format",
		dest="inputFormat",
	)
	parser.add_argument(
		"--write-format",
		dest="outputFormat",
		action="store",
	)

	parser.add_argument(
		"--direct",
		dest="direct",
		action="store_true",
		default=None,
		help="if possible, convert directly without loading into memory",
	)
	parser.add_argument(
		"--indirect",
		dest="direct",
		action="store_false",
		default=None,
		help=(
			"disable `direct` mode, load full data into memory before writing"
			", this is default"
		),
	)
	parser.add_argument(
		"--sqlite",
		dest="sqlite",
		action="store_true",
		default=None,
		help=(
			"use SQLite as middle storage instead of RAM in direct mode,"
			"for very large glossaries"
		),
	)
	parser.add_argument(
		"--no-sqlite",
		dest="sqlite",
		action="store_false",
		default=None,
		help="do not use SQLite mode",
	)

	parser.add_argument(
		"--no-progress-bar",
		dest="progressbar",
		action="store_false",
		default=None,
	)
	parser.add_argument(
		"--no-color",
		dest="noColor",
		action="store_true",
		default=not defaultHasColor,
	)

	parser.add_argument(
		"--sort",
		dest="sort",
		action="store_true",
		default=None,
	)
	parser.add_argument(
		"--no-sort",
		dest="sort",
		action="store_false",
		default=None,
	)
	parser.add_argument(
		"--sort-key",
		action="store",
		dest="sortKeyName",
		default=None,
		help="name of sort key",
	)
	parser.add_argument(
		"--sort-encoding",
		action="store",
		dest="sortEncoding",
		default=None,
		help="encoding of sort (default utf-8)",
	)

	# _______________________________

	parser.add_argument(
		"--source-lang",
		action="store",
		dest="sourceLang",
		default=None,
		help="source/query language",
	)
	parser.add_argument(
		"--target-lang",
		action="store",
		dest="targetLang",
		default=None,
		help="target/definition language",
	)
	parser.add_argument(
		"--name",
		action="store",
		dest="name",
		default=None,
		help="glossary name/title",
	)

	# _______________________________

	parser.add_argument(
		"--reverse",
		dest="reverse",
		action="store_true",
	)

	parser.add_argument(
		"inputFilename",
		action="store",
		default="",
		nargs="?",
	)
	parser.add_argument(
		"outputFilename",
		action="store",
		default="",
		nargs="?",
	)

	# _______________________________

	for key, option in UIBase.configDefDict.items():
		registerConfigOption(parser, key, option)


def validateFlags(args: "argparse.Namespace", log: logging.Logger) -> bool:
	from pyglossary.sort_keys import lookupSortKey, namedSortKeyList

	for param1, param2 in UIBase.conflictingParams:
		if getattr(args, param1) and getattr(args, param2):
			log.critical(
				"Conflicting flags: "
				f"--{param1.replace('_', '-')} and "
				f"--{param2.replace('_', '-')}",
			)
			return False

	if not args.sort:
		if args.sortKeyName:
			log.critical("Passed --sort-key without --sort")
			return False
		if args.sortEncoding:
			log.critical("Passed --sort-encoding without --sort")
			return False

	if args.sortKeyName and not lookupSortKey(args.sortKeyName):
		_valuesStr = ", ".join(_sk.name for _sk in namedSortKeyList)
		log.critical(
			f"Invalid sortKeyName={args.sortKeyName!r}"
			f". Supported values:\n{_valuesStr}",
		)
		return False

	return True


def configFromArgs(
	args: "argparse.Namespace",
	log: logging.Logger,
) -> dict[str, Any]:
	config = {}
	for key, option in UIBase.configDefDict.items():
		if not option.hasFlag:
			continue
		value = getattr(args, key, None)
		if value is None:
			continue
		log.debug(f"config: {key} = {value}")
		if not option.validate(value):
			log.error(f"invalid config value: {key} = {value!r}")
			continue
		config[key] = value
	return config




import typing
# -*- coding: utf-8 -*-

from typing import Generator, Iterator

from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.option import Option

__all__ = [
	"enable",
	"lname",
	"format",
	"description",
	"extensions",
	"extensionCreate",
	"singleFile",
	"kind",
	"wiki",
	"website",
	"optionsProp",
	"Reader",
	"Writer",
]

enable = False
lname = "testformat"
format = "Test"
description = "Test Format File(.test)"
extensions = (".test", ".tst")
extensionCreate = ".test"
singleFile = True
kind = "text"
wiki = ""
website = None

# key is option/argument name, value is instance of Option
optionsProp: "dict[str, Option]" = {}


class Reader:
	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._filename = ""
		self._wordCount = 0

	def __len__(self) -> int:
		# return the number of entries if you have it
		# if you don't, return 0 and progressbar will be disabled
		# self._wordCount can be set in self.open function
		# but if you want to set it, you should set it before
		# iteration begins and __iter__ method is called
		return self._wordCount

	def open(self, filename: str) -> None:
		# open the file, read headers / info and set info to self._glos
		# and set self._wordCount if you can
		# read-options should be keyword arguments in this method
		self._wordCount = 100
		# log.info(f"some useful message")
		# here read info from file and set to Glossary object
		self._glos.setInfo("name", "Test")
		desc = "Test glossary created by a PyGlossary plugin"
		self._glos.setInfo("description", desc)
		self._glos.setInfo("author", "Me")
		self._glos.setInfo("copyright", "GPL")

	def close(self) -> None:
		# this is called after reading/conversion is finished
		# if you have an open file object, close it here
		# if you need to clean up temp files, do it here
		pass

	def __iter__(self) -> "Iterator[EntryType]":
		# the easiest and simplest way to implement an Iterator is
		# by writing a generator, by calling: yield glos.newEntry(word, defi)
		# inside a loop (typically iterating over a file object for text file)
		# another way (which is harder) is by implementing __next__ method
		# and returning self in __iter__
		# that forces you to keep the state manually because __next__ is called
		# repeatedly, but __iter__ is only called once
		glos = self._glos
		for i in range(self._wordCount):
			# here get word and definition from file(depending on your format)
			word = f"word_{i}"
			defi = f"definition {i}"
			yield glos.newEntry(word, defi)


class Writer:
	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._filename = ""

	def open(self, filename: str) -> None:
		self._filename = filename

	def write(self) -> "Generator[None, EntryType, None]":
		glos = self._glos
		filename = self._filename  # noqa
		# log.info(f"some useful message")
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				# can save it with entry.save(directory)
				continue
			word = entry.s_word  # noqa
			defi = entry.defi  # noqa
			# here write word and defi to the output file (depending on
			# your format)
		# here read info from Glossaey object
		name = glos.getInfo("name")  # noqa
		desc = glos.getInfo("description")  # noqa
		author = glos.author  # noqa
		copyright = glos.getInfo("copyright")  # noqa
		# if an info key doesn't exist, getInfo returns empty string
		# now write info to the output file (depending on your output format)

	def finish(self) -> None:
		self._filename = ""



import typing
from collections.abc import Callable

__all__ = [
	"ErrorType",
	"LexType",
	"TitleLexType",
	"TitleTransformerType",
	"TransformerType",
]


class TransformerType(typing.Protocol):
	start: int
	pos: int
	input: str
	output: str
	currentKey: str
	attrs: "dict[str, str | None]"
	attrName: str
	audio: bool
	resFileSet: "set[str]"
	exampleColor: str

	def __init__(self) -> None:
		pass

	def end(self) -> bool:
		pass

	def move(self, chars: int) -> None:
		pass

	def next(self) -> str:
		pass

	def follows(self, st: str) -> bool:
		pass

	def skipAny(self, chars: str) -> None:
		pass

	def addText(self, st: str) -> None:
		pass

	def resetBuf(self) -> str:
		pass


class TitleTransformerType(TransformerType, typing.Protocol):
	title: str
	outputAlt: str

	def addText2(self, st: str) -> None:
		pass


ErrorType = str | None

# it is an State Function (state as in state machine)
LexType = Callable[[TransformerType], tuple["LexType", ErrorType]] | None

TitleLexType = Callable[[TitleTransformerType], tuple["TitleLexType", ErrorType]] | None



# -*- coding: utf-8 -*-
# mypy: ignore-errors

from collections.abc import Iterator
from io import BytesIO
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from pyglossary.glossary_types import (
		EntryType,
		GlossaryType,
	)
	from pyglossary.lxml_types import Element
	from pyglossary.option import Option
from pyglossary.compression import (
	compressionOpen,
	stdCompressions,
)
from pyglossary.core import log, pip
from pyglossary.html_utils import unescape_unicode

__all__ = [
	"Reader",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "iupac_goldbook"
format = "IUPACGoldbook"
description = "IUPAC goldbook (.xml)"
extensions = ()
extensionCreate = ".xml"
singleFile = True
kind = "text"
wiki = ""
website = "https://goldbook.iupac.org/"
optionsProp: "dict[str, Option]" = {}


class Reader:
	compressions = stdCompressions
	depends = {
		"lxml": "lxml",
	}

	def __init__(self, glos: "GlossaryType") -> None:
		self._glos = glos
		self._filename = ""
		self._file = None
		self._fileSize = 0
		self._termByCode = None

	def __len__(self) -> int:
		return 0

	def close(self) -> None:
		if self._file:
			self._file.close()
			self._file = None
		self._filename = ""
		self._fileSize = 0
		self._termByCode = None

	def open(self, filename: str) -> None:
		try:
			from lxml import etree as ET
		except ModuleNotFoundError as e:
			e.msg += f", run `{pip} install lxml` to install"
			raise e

		self._filename = filename
		_file = compressionOpen(filename, mode="rb")
		_file.seek(0, 2)
		self._fileSize = _file.tell()
		_file.seek(0)

		chunk = _file.read(800)
		chunk_end = chunk.find(b"<entries>")
		chunk = chunk[:chunk_end]
		chunk += b"</vocabulary>"

		infoRoot = ET.fromstring(chunk)
		self.setMetadata(infoRoot)

		_file.seek(0)
		context = ET.iterparse(
			_file,
			events=("end",),
			tag="entry",
		)
		termByCode = {}
		for _, elem in context:
			termE = elem.find("./term")
			if termE is None:
				continue
			term = self.getTerm(termE)
			codeE = elem.find("./code")
			if codeE is None:
				continue
			termByCode[codeE.text] = term
		self._termByCode = termByCode

		_file.close()

	def setGlosInfo(self, key: str, value: str) -> None:
		if value is None:
			return
		self._glos.setInfo(key, unescape_unicode(value))

	def setMetadata(self, header: "Element") -> None:
		if header is None:
			return

		title = header.find("./title")
		if title:
			self.setGlosInfo("name", title.text)

		publisher = header.find("./publisher")
		if publisher:
			self.setGlosInfo("publisher", publisher.text)

		isbn = header.find("./isbn")
		if isbn:
			self.setGlosInfo("isbn", isbn.text)

		doi = header.find("./doi")
		if doi:
			self.setGlosInfo("doi", doi.text)

		accessdate = header.find("./accessdate")
		if accessdate:
			self.setGlosInfo("creationTime", accessdate.text)

	@staticmethod
	def tostring(
		elem: "Element",
	) -> str:
		from lxml import etree as ET

		return (
			ET.tostring(
				elem,
				method="html",
				pretty_print=True,
			)
			.decode("utf-8")
			.strip()
		)

	@staticmethod
	def innerXML(elem: "Element") -> str:
		from lxml import etree as ET

		elemName = elem.xpath("name(/*)")
		resultStr = ""
		for e in elem.xpath("/" + elemName + "/node()"):
			if isinstance(e, str):
				pass  # resultStr += e
			else:
				resultStr += ET.tostring(e, encoding="unicode")

		return resultStr

	def getTerm(self, termE: "Element") -> str:  # noqa: PLR6301
		from lxml import etree as ET

		term = (
			ET.tostring(
				termE,
				method="html",
				pretty_print=False,
			)
			.decode("utf-8")
			.strip()[6:-7]
			.strip()
		)
		term = unescape_unicode(term)
		term = term.replace("<i>", "").replace("</i>", "")
		return term  # noqa: RET504

	def __iter__(self) -> "Iterator[EntryType]":  # noqa: PLR0912
		from lxml import etree as ET

		glos = self._glos
		fileSize = self._fileSize
		termByCode = self._termByCode

		self._file = _file = compressionOpen(self._filename, mode="rb")
		context = ET.iterparse(
			self._file,
			events=("end",),
			tag="entry",
		)
		for _, elem in context:  # noqa: PLR1702
			codeE = elem.find("./code")
			if codeE is None:
				continue
			code = codeE.text

			_id = elem.attrib.get("id")
			termE = elem.find("./term")
			if termE is None:
				log.warning(f"no term, {code=}, {_id=}")
				continue

			term = self.getTerm(termE)

			words = []
			if term:
				words.append(term)
			if code:
				words.append(code)

			# if _id is not None:
			# 	words.append(f"id{_id}")

			identifierTerm = elem.find("./identifiers/term")
			if identifierTerm is not None and identifierTerm.text:
				words.append(identifierTerm.text)

			identifierSynonym = elem.find("./identifiers/synonym")
			if identifierSynonym is not None and identifierSynonym.text:
				words.append(identifierSynonym.text)

			defiParts = []

			definition = elem.find("./definition")
			if definition is None or not definition.text:
				pass
			else:
				defiParts.append(definition.text)

			definitionEntryList = elem.findall("./definition/entry")
			if definitionEntryList:
				bio = BytesIO()
				with ET.htmlfile(bio, encoding="utf-8") as hf:
					with hf.element("ol"):
						for item in definitionEntryList:
							if not item.text:
								continue
							with hf.element("li"):
								hf.write(item.text)
				listHtml = bio.getvalue().decode("utf-8")
				defiParts.append(listHtml)

			replacedbyE = elem.find("./replacedby")
			if replacedbyE is not None:
				replacedby = replacedbyE.text
				replacedbyCode = replacedby.split(".")[-1]
				replacedbyTerm = termByCode.get(replacedbyCode)
				if replacedbyTerm is None:
					log.warning(f"{term}: {replacedby=}")
					replacedbyTerm = replacedbyCode
				defiParts.append(
					f'Replaced by: <a href="bword://{replacedbyTerm}">{replacedbyTerm}</a>',
				)

			relatedList = elem.findall("./related/entry")
			if relatedList:
				relatedLinkList = []
				for related in relatedList:
					relatedURL = related.text
					relatedCode = relatedURL.split("/")[-1]
					relatedTerm = termByCode.get(relatedCode)
					if not relatedTerm:
						log.warning(f"{term}: {relatedURL=}")
						relatedTerm = relatedCode
					relatedLinkList.append(
						f'<a href="bword://{relatedTerm}">{relatedTerm}</a>',
					)
				defiParts.append("Related: " + ", ".join(relatedLinkList))

			lastupdatedE = elem.find("./lastupdated")
			if lastupdatedE is not None:
				defiParts.append(f"Last updated: {lastupdatedE.text}")

			urlE = elem.find("./url")
			if urlE is not None:
				defiParts.append(f'<a href="{urlE.text}">More info.</a>')

			if len(defiParts) > 1:
				defiParts.insert(1, "")

			try:
				defi = "<br/>".join(defiParts)
			except Exception:
				log.error(f"{defiParts = }")
				continue

			yield glos.newEntry(
				words,
				defi,
				defiFormat="h",
				byteProgress=(_file.tell(), fileSize),
			)

			# clean up preceding siblings to save memory
			# this can reduce memory usage from >300 MB to ~25 MB
			parent = elem.getparent()
			if parent is None:
				continue
			while elem.getprevious() is not None:
				del parent[0]



import unittest

from pyglossary.plugins.babylon_bgl.bgl_reader_debug import isASCII


class BglInternalTest(unittest.TestCase):
	def test_isASCII(self):
		f = isASCII
		self.assertEqual(f(""), True)
		self.assertEqual(f("abc"), True)
		self.assertEqual(f("xyz"), True)
		self.assertEqual(f("ABC"), True)
		self.assertEqual(f("XYZ"), True)
		self.assertEqual(f("1234567890"), True)
		self.assertEqual(f("\n\r\t"), True)
		self.assertEqual(f("\x80"), False)
		self.assertEqual(f("abc\x80"), False)
		self.assertEqual(f("abc\xff"), False)



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2019 Saeed Rasooli <saeed.gnu@gmail.com>
# Copyright © 2007 Mehdi Bayazee <Bayazee@Gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/gpl.txt>.
# Also available in /usr/share/common-licenses/GPL on Debian systems
# or /usr/share/licenses/common/GPL3/license.txt on ArchLinux


# Gregorian calendar:
# http://en.wikipedia.org/wiki/Gregorian_calendar

from datetime import datetime

__all__ = ["isLeap", "jd_to", "to_jd"]

name = "gregorian"
desc = "Gregorian"

epoch = 1721426

options = ()


def save() -> None:
	pass


def isLeap(y: int) -> bool:
	return y % 4 == 0 and not (y % 100 == 0 and y % 400 != 0)


def to_jd(year: int, month: int, day: int) -> int:
	if 0 < year < 10000:  # > 1.5x faster # noqa: PLR2004
		return datetime(year, month, day).toordinal() + 1721425

	if month <= 2:  # noqa: PLR2004
		tm = 0
	elif isLeap(year):
		tm = -1
	else:
		tm = -2

	return (
		epoch - 1 +
		365 * (year - 1) +
		(year - 1) // 4 +
		-((year - 1) // 100) +
		(year - 1) // 400 +
		(367 * month - 362) // 12 +
		tm +
		day
	)


def jd_to(jd: int) -> "tuple[int, int, int]":
	ordinal = jd - 1721425
	if 0 < ordinal < 3652060:  # > 4x faster  # noqa: PLR2004
		# datetime(9999, 12, 31).toordinal() == 3652059
		dt = datetime.fromordinal(ordinal)
		return (dt.year, dt.month, dt.day)

	# wjd = floor(jd - 0.5) + 0.5
	qc, dqc = divmod(jd - epoch, 146097)  # qc ~~ quadricent
	cent, dcent = divmod(dqc, 36524)
	quad, dquad = divmod(dcent, 1461)
	yindex = dquad // 365  # divmod(dquad, 365)[0]
	year = (
		qc * 400 +
		cent * 100 +
		quad * 4 +
		yindex +
		(cent != 4 and yindex != 4)  # noqa: PLR2004
	)
	yearday = jd - to_jd(year, 1, 1)

	if jd < to_jd(year, 3, 1):
		leapadj = 0
	elif isLeap(year):
		leapadj = 1
	else:
		leapadj = 2

	month = ((yearday + leapadj) * 12 + 373) // 367
	day = jd - to_jd(year, month, 1) + 1
	return year, month, day



# -*- coding: utf-8 -*-
# mypy: ignore-errors
#
# Copyright © 2020 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.


from . import gtk
from .utils import (
	VBox,
	imageFromFile,
	pack,
)

__all__ = ["AboutWidget"]


class AboutTabTitleBox(gtk.Box):
	def __init__(self, title: str, icon: str) -> None:
		gtk.Box.__init__(self, orientation=gtk.Orientation.VERTICAL)
		self.set_spacing(10)
		pack(self, VBox(), expand=0)
		if icon:
			image = imageFromFile(icon)
			image.get_pixel_size()
			image.set_size_request(24, 24)
			# I don't know how to stop Gtk from resizing the image
			# I should probably use svg files to avoid blurry images
			pack(self, image, expand=0)
		if title:
			pack(self, gtk.Label(label=title), expand=0)
		pack(self, VBox(), expand=0)
		self.set_size_request(60, 60)

	# def do_get_preferred_height_for_width(self, size: int) -> "tuple[int, int]":
	# 	height = int(size * 1.5)
	# 	return height, height

	# returns: (minimum: int, natural: int,
	# 	minimum_baseline: int, natural_baseline: int)
	# def do_measure(self, orientation, for_size):
	# 	return (for_size, for_size, for_size, for_size)


class AboutWidget(gtk.Box):
	def __init__(  # noqa: PLR0913
		self,
		logo: str = "",
		header: str = "",
		about: str = "",
		authors: str = "",
		license_text: str = "",
		**_kwargs,
	) -> None:
		gtk.Box.__init__(self, orientation=gtk.Orientation.VERTICAL)
		self.set_spacing(15)
		##
		headerBox = gtk.Box(orientation=gtk.Orientation.HORIZONTAL)
		if logo:
			pack(headerBox, imageFromFile(logo))
		headerLabel = gtk.Label(label=header)
		headerLabel.set_selectable(True)
		pack(headerBox, headerLabel)
		headerBox.show()
		pack(self, headerBox)
		##
		notebook = gtk.Notebook()
		self.notebook = notebook
		pack(self, notebook, expand=True)
		notebook.set_tab_pos(gtk.PositionType.LEFT)
		##
		tab1_about = self.newTabLabelWidget(about)
		tab2_authors = self.newTabWidgetTextView(authors)
		tab3_license = self.newTabWidgetTextView(license_text)
		##
		tabs = [
			(tab1_about, self.newTabTitle("About", "dialog-information-22.png")),
			(tab2_authors, self.newTabTitle("Authors", "author-22.png")),
			(tab3_license, self.newTabTitle("License", "license-22.png")),
		]
		##
		for widget, titleW in tabs:
			notebook.append_page(widget, titleW)
		##
		self.show()

	# <a href="...">Something</a> does not work with TextView
	@staticmethod
	def newTabWidgetTextView(
		text: str,
		wrap: bool = False,
		justification: "gtk.Justification | None" = None,
	):
		tv = gtk.TextView()
		if wrap:
			tv.set_wrap_mode(gtk.WrapMode.WORD)
		if justification is not None:
			tv.set_justification(justification)
		tv.set_cursor_visible(False)
		# tv.set_border_width(10)
		buf = tv.get_buffer()
		# buf.insert_markup(buf.get_end_iter(), markup=text,
		# len=len(text.encode("utf-8")))
		buf.set_text(text)
		tv.show()
		swin = gtk.ScrolledWindow()
		swin.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
		# swin.set_border_width(0)
		swin.set_child(tv)
		return swin

	@staticmethod
	def newTabLabelWidget(
		text: str,
		# wrap: bool = False,
		# justification: "gtk.Justification | None" = None,
	):
		box = VBox()
		# box.set_border_width(10)
		label = gtk.Label()
		label.set_selectable(True)
		label.set_xalign(0)
		label.set_yalign(0)
		pack(box, label, 0, 0)
		# if wrap:
		# 	tv.set_wrap_mode(gtk.WrapMode.WORD)
		# if justification is not None:
		# 	tv.set_justification(justification)
		# label.set_cursor_visible(False)
		# label.set_border_width(10)
		label.set_markup(text)
		label.show()
		swin = gtk.ScrolledWindow()
		swin.set_policy(gtk.PolicyType.AUTOMATIC, gtk.PolicyType.AUTOMATIC)
		# swin.set_border_width(0)
		swin.set_child(box)
		return swin

	@staticmethod
	def newTabTitle(title: str, icon: str):
		return AboutTabTitleBox(title, icon)



import io
from typing import TYPE_CHECKING

if TYPE_CHECKING:
	from collections.abc import Iterator

__all__ = ["nullBinaryIO", "nullTextIO"]


class _NullBinaryIO(io.BufferedIOBase):  # noqa: PLR0904
	def __enter__(self, *args):
		raise NotImplementedError

	def __exit__(self, *args):
		raise NotImplementedError

	def close(self) -> None:
		pass

	def fileno(self) -> int:
		raise NotImplementedError

	def flush(self) -> None:
		raise NotImplementedError

	def isatty(self) -> bool:
		raise NotImplementedError

	def readable(self) -> bool:
		raise NotImplementedError

	def seek(self, pos: int, whence: int = 0) -> int:
		raise NotImplementedError

	def seekable(self) -> bool:
		raise NotImplementedError

	def tell(self) -> int:
		raise NotImplementedError

	def truncate(self, pos: "int | None" = None) -> int:
		raise NotImplementedError

	def writable(self) -> bool:
		raise NotImplementedError

	def detach(self) -> "io.RawIOBase":
		raise NotImplementedError

	def read(self, n: "int | None" = None) -> bytes:
		raise NotImplementedError

	def read1(self, n: "int | None" = None) -> bytes:
		raise NotImplementedError

	def readinto(self, buffer) -> int:
		raise NotImplementedError

	def readinto1(self, buffer) -> int:
		raise NotImplementedError

	# data: "bytearray|memoryview|array[Any]|io.mmap|io._CData|io.PickleBuffer"
	def write(self, data: bytes) -> int:  # type: ignore
		raise NotImplementedError

	def __iter__(self) -> "Iterator[bytes]":
		raise NotImplementedError

	def __next__(self) -> bytes:
		raise NotImplementedError

	def readline(self, size: "int | None" = -1) -> bytes:
		raise NotImplementedError

	def readlines(self, hint: int = -1) -> "list[bytes]":
		raise NotImplementedError

	def writelines(self, lines: "list[bytes]") -> None:  # type: ignore
		raise NotImplementedError


class _NullTextIO(io.TextIOBase):  # noqa: PLR0904
	def __enter__(self, *args):
		raise NotImplementedError

	def __exit__(self, *args):
		raise NotImplementedError

	def close(self) -> None:
		pass

	def fileno(self) -> int:
		raise NotImplementedError

	def flush(self) -> None:
		raise NotImplementedError

	def isatty(self) -> bool:
		raise NotImplementedError

	def readable(self) -> bool:
		raise NotImplementedError

	def seek(self, pos: int, whence: int = 0) -> int:
		raise NotImplementedError

	def seekable(self) -> bool:
		raise NotImplementedError

	def tell(self) -> int:
		raise NotImplementedError

	def truncate(self, pos: "int | None" = None) -> int:
		raise NotImplementedError

	def writable(self) -> bool:
		raise NotImplementedError

	def detach(self) -> "io.IOBase":  # type: ignore
		raise NotImplementedError

	def read(self, n: "int | None" = None) -> str:
		raise NotImplementedError

	def read1(self, n: "int | None" = None) -> str:
		raise NotImplementedError

	def readinto(self, buffer) -> "io.BufferedIOBase":
		raise NotImplementedError

	def readinto1(self, buffer) -> "io.BufferedIOBase":
		raise NotImplementedError

	# data: "bytearray|memoryview|array[Any]|io.mmap|io._CData|io.PickleBuffer"
	def write(self, data: "bytes") -> int:  # type: ignore
		raise NotImplementedError

	def __iter__(self) -> "Iterator[str]":  # type: ignore
		raise NotImplementedError

	def __next__(self) -> str:  # type: ignore
		raise NotImplementedError

	def readline(self, size: "int | None" = -1) -> str:  # type: ignore
		raise NotImplementedError

	def readlines(self, hint: int = -1) -> "list[str]":  # type: ignore
		raise NotImplementedError

	def writelines(self, lines: "list[str]") -> None:  # type: ignore
		raise NotImplementedError


nullBinaryIO = _NullBinaryIO()
nullTextIO = _NullTextIO()



# -*- coding: utf-8 -*-
# mypy: ignore-errors
#
# Copyright © 2016-2017 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# You can get a copy of GNU General Public License along this program
# But you can always get it from http://www.gnu.org/licenses/gpl.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.


from gi.repository import Gdk as gdk

from .utils import gtk_event_iteration_loop


class MyDialog:
	def startWaiting(self):
		self.queue_draw()
		self.vbox.set_sensitive(False)
		self.get_window().set_cursor(gdk.Cursor.new(gdk.CursorType.WATCH))
		gtk_event_iteration_loop()

	def endWaiting(self):
		self.get_window().set_cursor(gdk.Cursor.new(gdk.CursorType.LEFT_PTR))
		self.vbox.set_sensitive(True)

	def waitingDo(self, func, *args, **kwargs):
		self.startWaiting()
		try:
			func(*args, **kwargs)
		except Exception as e:
			raise e
		finally:
			self.endWaiting()



# coding: utf-8
# mypy: ignore-errors

# Copyright (C) 2016-2023 Saeed Rasooli on https://github.com/ilius/pyglossary/
# Copyright (C) 2015 Z. H. Liu on https://github.com/zhansliu/writemdict

# pureSalsa20.py -- a pure Python implementation of the Salsa20 cipher,
# ported to Python 3

# v4.0: Added Python 3 support, dropped support for Python <= 2.5.

# // zhansliu

# Original comments below.

# ====================================================================
# There are comments here by two authors about three pieces of software:
# 	comments by Larry Bugbee about
# 		Salsa20, the stream cipher by Daniel J. Bernstein
# 				(including comments about the speed of the C version) and
# 		pySalsa20, Bugbee's own Python wrapper for salsa20.c
# 				(including some references), and
# 	comments by Steve Witham about
# 		pureSalsa20, Witham's pure Python 2.5 implementation of Salsa20,
# 			which follows pySalsa20's API, and is in this file.

# Salsa20: a Fast Streaming Cipher (comments by Larry Bugbee)
# -----------------------------------------------------------

# Salsa20 is a fast stream cipher written by Daniel Bernstein
# that basically uses a hash function and XOR making for fast
# encryption.  (Decryption uses the same function.)  Salsa20
# is simple and quick.

# Some Salsa20 parameter values...
# 	design strength    128 bits
# 	key length         128 or 256 bits, exactly
# 	IV, aka nonce      64 bits, always
# 	chunk size         must be in multiples of 64 bytes

# Salsa20 has two reduced versions, 8 and 12 rounds each.

# One benchmark (10 MB):
# 	1.5GHz PPC G4     102/97/89 MB/sec for 8/12/20 rounds
# 	AMD Athlon 2500+   77/67/53 MB/sec for 8/12/20 rounds
# 		(no I/O and before Python GC kicks in)

# Salsa20 is a Phase 3 finalist in the EU eSTREAM competition
# and appears to be one of the fastest ciphers.  It is well
# documented so I will not attempt any injustice here.  Please
# see "References" below.

# ...and Salsa20 is "free for any use".


# pySalsa20: a Python wrapper for Salsa20 (Comments by Larry Bugbee)
# ------------------------------------------------------------------

# pySalsa20.py is a simple ctypes Python wrapper.  Salsa20 is
# as it's name implies, 20 rounds, but there are two reduced
# versions, 8 and 12 rounds each.  Because the APIs are
# identical, pySalsa20 is capable of wrapping all three
# versions (number of rounds hardcoded), including a special
# version that allows you to set the number of rounds with a
# set_rounds() function.  Compile the version of your choice
# as a shared library (not as a Python extension), name and
# install it as libsalsa20.so.

# Sample usage:
# 	from pySalsa20 import Salsa20
# 	s20 = Salsa20(key, IV)
# 	dataout = s20.encryptBytes(datain)   # same for decrypt

# This is EXPERIMENTAL software and intended for educational
# purposes only.  To make experimentation less cumbersome,
# pySalsa20 is also free for any use.

# THIS PROGRAM IS PROVIDED WITHOUT WARRANTY OR GUARANTEE OF
# ANY KIND.  USE AT YOUR OWN RISK.

# Enjoy,

# Larry Bugbee
# bugbee@seanet.com
# April 2007


# References:
# -----------
# 	http://en.wikipedia.org/wiki/Salsa20
# 	http://en.wikipedia.org/wiki/Daniel_Bernstein
# 	http://cr.yp.to/djb.html
# 	http://www.ecrypt.eu.org/stream/salsa20p3.html
# 	http://www.ecrypt.eu.org/stream/p3ciphers/salsa20/salsa20_p3source.zip


# Prerequisites for pySalsa20:
# ----------------------------
# 	- Python 2.5 (haven't tested in 2.4)


# pureSalsa20: Salsa20 in pure Python 2.5 (comments by Steve Witham)
# ------------------------------------------------------------------

# pureSalsa20 is the stand-alone Python code in this file.
# It implements the underlying Salsa20 core algorithm
# and emulates pySalsa20's Salsa20 class API (minus a bug(*)).

# pureSalsa20 is MUCH slower than libsalsa20.so wrapped with pySalsa20--
# about 1/1000 the speed for Salsa20/20 and 1/500 the speed for Salsa20/8,
# when encrypting 64k-byte blocks on my computer.

# pureSalsa20 is for cases where portability is much more important than
# speed.  I wrote it for use in a "structured" random number generator.

# There are comments about the reasons for this slowness in
# 		http://www.tiac.net/~sw/2010/02/PureSalsa20

# Sample usage:
# 	from pureSalsa20 import Salsa20
# 	s20 = Salsa20(key, IV)
# 	dataout = s20.encryptBytes(datain)   # same for decrypt

# I took the test code from pySalsa20, added a bunch of tests including
# rough speed tests, and moved them into the file testSalsa20.py.
# To test both pySalsa20 and pureSalsa20, type
# 	python testSalsa20.py

# (*)The bug (?) in pySalsa20 is this.  The rounds variable is global to the
# libsalsa20.so library and not switched when switching between instances
# of the Salsa20 class.
# 	s1 = Salsa20( key, IV, 20 )
# 	s2 = Salsa20( key, IV, 8 )
# In this example,
# 	with pySalsa20, both s1 and s2 will do 8 rounds of encryption.
# 	with pureSalsa20, s1 will do 20 rounds and s2 will do 8 rounds.
# Perhaps giving each instance its own nRounds variable, which
# is passed to the salsa20wordtobyte() function, is insecure.  I'm not a
# cryptographer.

# pureSalsa20.py and testSalsa20.py are EXPERIMENTAL software and
# intended for educational purposes only.  To make experimentation less
# cumbersome, pureSalsa20.py and testSalsa20.py are free for any use.

# Revisions:
# ----------
# 	p3.2   Fixed bug that initialized the output buffer with plaintext!
# 			Saner ramping of nreps in speed test.
# 			Minor changes and print statements.
# 	p3.1   Took timing variability out of add32() and rot32().
# 			Made the internals more like pySalsa20/libsalsa .
# 			Put the semicolons back in the main loop!
# 			In encryptBytes(), modify a byte array instead of appending.
# 			Fixed speed calculation bug.
# 			Used subclasses instead of patches in testSalsa20.py .
# 			Added 64k-byte messages to speed test to be fair to pySalsa20.
# 	p3     First version, intended to parallel pySalsa20 version 3.

# More references:
# ----------------
# 	http://www.seanet.com/~bugbee/crypto/salsa20/          [pySalsa20]
# 	http://cr.yp.to/snuffle.html        [The original name of Salsa20]
# 	http://cr.yp.to/snuffle/salsafamily-20071225.pdf [ Salsa20 design]
# 	http://www.tiac.net/~sw/2010/02/PureSalsa20

# THIS PROGRAM IS PROVIDED WITHOUT WARRANTY OR GUARANTEE OF
# ANY KIND.  USE AT YOUR OWN RISK.

# Cheers,

# Steve Witham sw at remove-this tiac dot net
# February, 2010


import operator
from struct import Struct

__all__ = ["Salsa20"]

little_u64 = Struct("<Q")  # little-endian 64-bit unsigned.
# Unpacks to a tuple of one element!

little16_i32 = Struct("<16i")  # 16 little-endian 32-bit signed ints.
little4_i32 = Struct("<4i")  # 4 little-endian 32-bit signed ints.
little2_i32 = Struct("<2i")  # 2 little-endian 32-bit signed ints.

_version = "p4.0"

# ----------- Salsa20 class which emulates pySalsa20.Salsa20 ---------------


class Salsa20:
	def __init__(self, key=None, IV=None, rounds=20) -> None:
		self._lastChunk64 = True
		self._IVbitlen = 64  # must be 64 bits
		self.ctx = [0] * 16
		if key:
			self.setKey(key)
		if IV:
			self.setIV(IV)

		self.setRounds(rounds)

	def setKey(self, key):
		assert isinstance(key, bytes)
		ctx = self.ctx
		if len(key) == 32:  # recommended
			constants = b"expand 32-byte k"
			ctx[1], ctx[2], ctx[3], ctx[4] = little4_i32.unpack(key[0:16])
			ctx[11], ctx[12], ctx[13], ctx[14] = little4_i32.unpack(key[16:32])
		elif len(key) == 16:
			constants = b"expand 16-byte k"
			ctx[1], ctx[2], ctx[3], ctx[4] = little4_i32.unpack(key[0:16])
			ctx[11], ctx[12], ctx[13], ctx[14] = little4_i32.unpack(key[0:16])
		else:
			raise ValueError("key length isn't 32 or 16 bytes.")
		ctx[0], ctx[5], ctx[10], ctx[15] = little4_i32.unpack(constants)

	def setIV(self, IV):
		assert isinstance(IV, bytes)
		assert len(IV) * 8 == 64, "nonce (IV) not 64 bits"
		self.IV = IV
		ctx = self.ctx
		ctx[6], ctx[7] = little2_i32.unpack(IV)
		ctx[8], ctx[9] = 0, 0  # Reset the block counter.

	setNonce = setIV  # support an alternate name

	def setCounter(self, counter):
		assert isinstance(counter, int)
		assert 0 <= counter < 1 << 64, "counter < 0 or >= 2**64"
		ctx = self.ctx
		ctx[8], ctx[9] = little2_i32.unpack(little_u64.pack(counter))

	def getCounter(self):
		return little_u64.unpack(little2_i32.pack(*self.ctx[8:10]))[0]

	def setRounds(self, rounds, testing=False):
		assert testing or rounds in {8, 12, 20}, "rounds must be 8, 12, 20"
		self.rounds = rounds

	def encryptBytes(self, data: bytes) -> bytes:
		assert isinstance(data, bytes), "data must be byte string"
		assert self._lastChunk64, "previous chunk not multiple of 64 bytes"
		lendata = len(data)
		munged = bytearray(lendata)
		for i in range(0, lendata, 64):
			h = salsa20_wordtobyte(self.ctx, self.rounds, checkRounds=False)
			self.setCounter((self.getCounter() + 1) % 2**64)
			# Stopping at 2^70 bytes per nonce is user's responsibility.
			for j in range(min(64, lendata - i)):
				munged[i + j] = data[i + j] ^ h[j]

		self._lastChunk64 = not lendata % 64
		return bytes(munged)

	decryptBytes = encryptBytes  # encrypt and decrypt use same function


# --------------------------------------------------------------------------


def salsa20_wordtobyte(input_, nRounds=20, checkRounds=True):
	"""
	Do nRounds Salsa20 rounds on a copy of
			input: list or tuple of 16 ints treated as little-endian unsigneds.
	Returns a 64-byte string.
	"""
	assert isinstance(input_, list | tuple) and len(input_) == 16
	assert not checkRounds or nRounds in {8, 12, 20}

	x = list(input_)

	XOR = operator.xor

	ROTATE = rot32
	PLUS = add32

	for _ in range(nRounds // 2):
		# These ...XOR...ROTATE...PLUS... lines are from ecrypt-linux.c
		# unchanged except for indents and the blank line between rounds:
		x[4] = XOR(x[4], ROTATE(PLUS(x[0], x[12]), 7))
		x[8] = XOR(x[8], ROTATE(PLUS(x[4], x[0]), 9))
		x[12] = XOR(x[12], ROTATE(PLUS(x[8], x[4]), 13))
		x[0] = XOR(x[0], ROTATE(PLUS(x[12], x[8]), 18))
		x[9] = XOR(x[9], ROTATE(PLUS(x[5], x[1]), 7))
		x[13] = XOR(x[13], ROTATE(PLUS(x[9], x[5]), 9))
		x[1] = XOR(x[1], ROTATE(PLUS(x[13], x[9]), 13))
		x[5] = XOR(x[5], ROTATE(PLUS(x[1], x[13]), 18))
		x[14] = XOR(x[14], ROTATE(PLUS(x[10], x[6]), 7))
		x[2] = XOR(x[2], ROTATE(PLUS(x[14], x[10]), 9))
		x[6] = XOR(x[6], ROTATE(PLUS(x[2], x[14]), 13))
		x[10] = XOR(x[10], ROTATE(PLUS(x[6], x[2]), 18))
		x[3] = XOR(x[3], ROTATE(PLUS(x[15], x[11]), 7))
		x[7] = XOR(x[7], ROTATE(PLUS(x[3], x[15]), 9))
		x[11] = XOR(x[11], ROTATE(PLUS(x[7], x[3]), 13))
		x[15] = XOR(x[15], ROTATE(PLUS(x[11], x[7]), 18))

		x[1] = XOR(x[1], ROTATE(PLUS(x[0], x[3]), 7))
		x[2] = XOR(x[2], ROTATE(PLUS(x[1], x[0]), 9))
		x[3] = XOR(x[3], ROTATE(PLUS(x[2], x[1]), 13))
		x[0] = XOR(x[0], ROTATE(PLUS(x[3], x[2]), 18))
		x[6] = XOR(x[6], ROTATE(PLUS(x[5], x[4]), 7))
		x[7] = XOR(x[7], ROTATE(PLUS(x[6], x[5]), 9))
		x[4] = XOR(x[4], ROTATE(PLUS(x[7], x[6]), 13))
		x[5] = XOR(x[5], ROTATE(PLUS(x[4], x[7]), 18))
		x[11] = XOR(x[11], ROTATE(PLUS(x[10], x[9]), 7))
		x[8] = XOR(x[8], ROTATE(PLUS(x[11], x[10]), 9))
		x[9] = XOR(x[9], ROTATE(PLUS(x[8], x[11]), 13))
		x[10] = XOR(x[10], ROTATE(PLUS(x[9], x[8]), 18))
		x[12] = XOR(x[12], ROTATE(PLUS(x[15], x[14]), 7))
		x[13] = XOR(x[13], ROTATE(PLUS(x[12], x[15]), 9))
		x[14] = XOR(x[14], ROTATE(PLUS(x[13], x[12]), 13))
		x[15] = XOR(x[15], ROTATE(PLUS(x[14], x[13]), 18))

	for idx, item in enumerate(input_):
		x[idx] = PLUS(x[idx], item)
	return little16_i32.pack(*x)


# --------------------------- 32-bit ops -------------------------------


def trunc32(w):
	"""
	Return the bottom 32 bits of w as a Python int.
	This creates longs temporarily, but returns an int.
	"""
	w = int((w & 0x7FFFFFFF) | -(w & 0x80000000))
	assert isinstance(w, int)
	return w


def add32(a, b):
	"""
	Add two 32-bit words discarding carry above 32nd bit,
	and without creating a Python long.
	Timing shouldn't vary.
	"""
	lo = (a & 0xFFFF) + (b & 0xFFFF)
	hi = (a >> 16) + (b >> 16) + (lo >> 16)
	return (-(hi & 0x8000) | (hi & 0x7FFF)) << 16 | (lo & 0xFFFF)


def rot32(w, nLeft):
	"""
	Rotate 32-bit word left by nLeft or right by -nLeft
	without creating a Python long.
	Timing depends on nLeft but not on w.
	"""
	nLeft &= 31  # which makes nLeft >= 0
	if nLeft == 0:
		return w

	# Note: now 1 <= nLeft <= 31.
	# 	 RRRsLLLLLL   There are nLeft RRR's, (31-nLeft) LLLLLL's,
	# =>  sLLLLLLRRR   and one s which becomes the sign bit.
	RRR = ((w >> 1) & 0x7FFFFFFF) >> (31 - nLeft)
	sLLLLLL = -((1 << (31 - nLeft)) & w) | (0x7FFFFFFF >> nLeft) & w
	return RRR | (sLLLLLL << nLeft)


# --------------------------------- end -----------------------------------



# -*- coding: utf-8 -*-
#
# Copyright © 2008-2020 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
# Copyright © 2011-2012 kubtek <kubtek@gmail.com>
# This file is part of PyGlossary project, http://github.com/ilius/pyglossary
# Thanks to Raul Fernandes <rgfbr@yahoo.com.br> and Karl Grill for reverse
# engineering as part of https://sourceforge.net/projects/ktranslator/
#
# This program is a free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program. Or on Debian systems, from /usr/share/common-licenses/GPL
# If not, see <http://www.gnu.org/licenses/gpl.txt>.

"""
language properties.

In this short note we describe how Babylon select encoding for key words,
alternates and definitions.
There are source and target encodings. The source encoding is used to
encode keys and alternates, the target encoding is used to encode
definitions.
The source encoding is selected based on the source language of the
dictionary, the target encoding is tied to the target language.
Babylon Glossary Builder allows you to specify source and target languages.
If you open a Builder project (a file with .gpr extension) in a text
editor, you should find the following elements:
<bab:SourceCharset>Latin</bab:SourceCharset>
<bab:SourceLanguage>English</bab:SourceLanguage>
<bab:TargetCharset>Latin</bab:TargetCharset>
<bab:TargetLanguage>English</bab:TargetLanguage>
Here bab:SourceLanguage is the source language that you select in the
builder wizard, bab:SourceCharset - is the corresponding charset.
bab:TargetLanguage - target language, bab:TargetCharset - corresponding
charset.
Unfortunately, builder does not tell us what encoding corresponds to charset,
but we can detect it.

A few words about how definitions are encoded. If all chars of the
definition fall into the target charset, Babylon use that charset to encode
the definition. If at least one char does not fall into the target charset,
Babylon use utf-8 encoding, wrapping the definition into <charset c=U> and
</charset> tags.
You can make Babylon use utf-8 encoding for the whole dictionary, in that case
all definitions, keys and alternates are encoded with utf-8. See Babylon
Glossary Builder wizard, Glossary Properties tab, Advanced button, Use UTF-8
encoding check box. Definitions are not augmented with extra mackup in this
case, that is you'll not find charset tags in definitions.

How you can tell what encoding was used for the particular definition in
.bgl file? You need to check the following conditions.

Block type 3, code 0x11. If 0x8000 bit is set, the whole dictionary use
utf-8 encoding.

If the definition starts with <charset c=U>, that definition uses utf-8
encoding.

Otherwise you need to consult the target encoding.

Block type 3, code 0x1b. That field normally contains 1 byte code of the
target encoding. Codes fill the range of 0x41 to 0x4e. Babylon Builder
generate codes 0x42 - 0x4e. How to generate code 0x41?
Occasionally you may encounter the field value is four zero bytes. In this
case, I guess, the default encoding for the target language is used.

Block type 3, code 0x08. That field contains 4-bytes code of the target
language. The first three bytes are always zero, the last byte is the code.
Playing with Babylon Glossary builder we can find language codes corresponding
to target language. The language codes fill the range of 0 to 0x3d.

How to detect the target encoding? Here is the technique I've used.
- Create a babylon glossary source file ( a file with .gls extension) with
	the following contents. Start the file with utf-8 BOM for the builder
	to recognize the utf-8 encoding. Use unicode code point code as key,
	and a single unicode chars encoding in utf-8 as definition. Create keys
	for all code points in the range 32 - 0x10000, or you may use wider range.
	We do not use code points in the range 0-31, since they are control chars.
	You should skip the following three chars: & < >. Since the definition
	is supposed to contain html, these chars are be replaced by &amp; &lt;
	&gt; respectively. You should skip the char $ as well, it has special
	meaning in definitions (?). Skip all code point that cannot encoded in
	utf-8 (not all code points in the range 32-0x10000 represent valid chars).
- Now that you have a glossary source file, process it with builder selecting
	the desired target language. Make sure the "use utf-8" option is no set.
	You'll get a .bgl file.
- Process the generated .bgl file with pyglossary. Skip all definitions that
	start with <charset c=U> tag. Try to decode definitions using different
	encodings and match the result with the real value (key - code point char
	code). Thus you'll find the encoding having the best match.

	For example, you may do the following.
	Loop over all available encodings, loop over all definitions in the
	dictionary. Count the number of definitions that does not start with
	charset tag - total. Among them count the number of definitions that were
	correctly decoded - success. The encoding where total == success, is
	the target encoding.

There are a few problems I encountered. It looks like python does not
correctly implement cp932 and cp950 encodings. For Japanese charset I
got 99.12% match, and for Traditional Chinese charset I got even less -
66.97%. To conform my guess that Japanese is cp932 and Traditional Chinese
is cp950 I built a C++ utility that worked on the data extracted from .bgl
dictionary. I used WideCharToMultiByte function for conversion. The C++
utility confirmed the cp932 and cp950 encodings, I got 100% match.
"""

from dataclasses import dataclass

__all__ = ["BabylonLanguage", "languageByCode"]


@dataclass(slots=True, frozen=True)
class BabylonLanguage:

	"""
	Babylon language properties.

	name: bab:SourceLanguage, bab:TargetLanguage .gpr tags (English, French, Japanese)
	charset: bab:SourceCharset, bab:TargetCharset .gpr tags (Latin, Arabic, Cyrillic)
	encoding: Windows code page (cp1250, cp1251, cp1252)
	code: value of the type 3, code in .bgl file
	"""

	name: str
	charset: str
	encoding: str
	code: int
	code2: str = ""
	name2: str = ""


languages = (
	BabylonLanguage(
		name="English",
		charset="Latin",
		encoding="cp1252",
		code=0x00,
		code2="en",
	),
	BabylonLanguage(
		name="French",
		charset="Latin",
		encoding="cp1252",
		code=0x01,
		code2="fr",
	),
	BabylonLanguage(
		name="Italian",
		charset="Latin",
		encoding="cp1252",
		code=0x02,
		code2="it",
	),
	BabylonLanguage(
		name="Spanish",
		charset="Latin",
		encoding="cp1252",
		code=0x03,
		code2="es",
	),
	BabylonLanguage(
		name="Dutch",
		charset="Latin",
		encoding="cp1252",
		code=0x04,
		code2="nl",
	),
	BabylonLanguage(
		name="Portuguese",
		charset="Latin",
		encoding="cp1252",
		code=0x05,
		code2="pt",
	),
	BabylonLanguage(
		name="German",
		charset="Latin",
		encoding="cp1252",
		code=0x06,
		code2="de",
	),
	BabylonLanguage(
		name="Russian",
		charset="Cyrillic",
		encoding="cp1251",
		code=0x07,
		code2="ru",
	),
	BabylonLanguage(
		name="Japanese",
		charset="Japanese",
		encoding="cp932",
		code=0x08,
		code2="ja",
	),
	BabylonLanguage(
		name="Chinese",
		name2="Traditional Chinese",
		charset="Traditional Chinese",
		encoding="cp950",
		code=0x09,
		code2="zh",
	),
	BabylonLanguage(
		name="Chinese",
		name2="Simplified Chinese",
		charset="Simplified Chinese",
		encoding="cp936",
		code=0x0A,
		code2="zh",
	),
	BabylonLanguage(
		name="Greek",
		charset="Greek",
		encoding="cp1253",
		code=0x0B,
		code2="el",
	),
	BabylonLanguage(
		name="Korean",
		charset="Korean",
		encoding="cp949",
		code=0x0C,
		code2="ko",
	),
	BabylonLanguage(
		name="Turkish",
		charset="Turkish",
		encoding="cp1254",
		code=0x0D,
		code2="tr",
	),
	BabylonLanguage(
		name="Hebrew",
		charset="Hebrew",
		encoding="cp1255",
		code=0x0E,
		code2="he",
	),
	BabylonLanguage(
		name="Arabic",
		charset="Arabic",
		encoding="cp1256",
		code=0x0F,
		code2="ar",
	),
	BabylonLanguage(
		name="Thai",
		charset="Thai",
		encoding="cp874",
		code=0x10,
		code2="th",
	),
	BabylonLanguage(
		name="Other",
		charset="Latin",
		encoding="cp1252",
		code=0x11,
		code2="",  # none
	),
	BabylonLanguage(
		name="Chinese",
		name2="Other Simplified Chinese dialects",
		charset="Simplified Chinese",
		encoding="cp936",
		code=0x12,
		code2="zh",  # duplicate
	),
	BabylonLanguage(
		name="Chinese",
		name2="Other Traditional Chinese dialects",
		charset="Traditional Chinese",
		encoding="cp950",
		code=0x13,
		code2="zh",  # duplicate
	),
	BabylonLanguage(
		name="Other Eastern-European languages",
		charset="Eastern European",
		encoding="cp1250",
		code=0x14,
		code2="",  # none
	),
	BabylonLanguage(
		name="Other Western-European languages",
		charset="Latin",
		encoding="cp1252",
		code=0x15,
		code2="",  # none
	),
	BabylonLanguage(
		name="Other Russian languages",
		charset="Cyrillic",
		encoding="cp1251",
		code=0x16,
		code2="",  # none
	),
	BabylonLanguage(
		name="Other Japanese languages",
		charset="Japanese",
		encoding="cp932",
		code=0x17,
		code2="",  # none
	),
	BabylonLanguage(
		name="Other Baltic languages",
		charset="Baltic",
		encoding="cp1257",
		code=0x18,
		code2="bat",  # no 2-letter code
	),
	BabylonLanguage(
		name="Other Greek languages",
		charset="Greek",
		encoding="cp1253",
		code=0x19,
		code2="",  # none
	),
	BabylonLanguage(
		name="Other Korean dialects",
		charset="Korean",
		encoding="cp949",
		code=0x1A,
		code2="",  # none
	),
	BabylonLanguage(
		name="Other Turkish dialects",
		charset="Turkish",
		encoding="cp1254",
		code=0x1B,
		code2="",  # none
	),
	BabylonLanguage(
		name="Other Thai dialects",
		charset="Thai",
		encoding="cp874",
		code=0x1C,
		code2="tai",  # no 2-letter code, and "tha" / "th" is for "Thai"
	),
	BabylonLanguage(
		name="Polish",
		charset="Eastern European",
		encoding="cp1250",
		code=0x1D,
		code2="pl",
	),
	BabylonLanguage(
		name="Hungarian",
		charset="Eastern European",
		encoding="cp1250",
		code=0x1E,
		code2="hu",
	),
	BabylonLanguage(
		name="Czech",
		charset="Eastern European",
		encoding="cp1250",
		code=0x1F,
		code2="cs",
	),
	BabylonLanguage(
		name="Lithuanian",
		charset="Baltic",
		encoding="cp1257",
		code=0x20,
		code2="lt",
	),
	BabylonLanguage(
		name="Latvian",
		charset="Baltic",
		encoding="cp1257",
		code=0x21,
		code2="lv",
	),
	BabylonLanguage(
		name="Catalan",
		charset="Latin",
		encoding="cp1252",
		code=0x22,
		code2="ca",
	),
	BabylonLanguage(
		name="Croatian",
		charset="Eastern European",
		encoding="cp1250",
		code=0x23,
		code2="hr",
	),
	BabylonLanguage(
		name="Serbian",
		charset="Eastern European",
		encoding="cp1250",
		code=0x24,
		code2="sr",
	),
	BabylonLanguage(
		name="Slovak",
		charset="Eastern European",
		encoding="cp1250",
		code=0x25,
		code2="sk",
	),
	BabylonLanguage(
		name="Albanian",
		charset="Latin",
		encoding="cp1252",
		code=0x26,
		code2="sq",
	),
	BabylonLanguage(
		name="Urdu",
		charset="Arabic",
		encoding="cp1256",
		code=0x27,
		code2="ur",
	),
	BabylonLanguage(
		name="Slovenian",
		charset="Eastern European",
		encoding="cp1250",
		code=0x28,
		code2="sl",
	),
	BabylonLanguage(
		name="Estonian",
		charset="Latin",
		encoding="cp1252",
		code=0x29,
		code2="et",
	),
	BabylonLanguage(
		name="Bulgarian",
		charset="Eastern European",
		encoding="cp1250",
		code=0x2A,
		code2="bg",
	),
	BabylonLanguage(
		name="Danish",
		charset="Latin",
		encoding="cp1252",
		code=0x2B,
		code2="da",
	),
	BabylonLanguage(
		name="Finnish",
		charset="Latin",
		encoding="cp1252",
		code=0x2C,
		code2="fi",
	),
	BabylonLanguage(
		name="Icelandic",
		charset="Latin",
		encoding="cp1252",
		code=0x2D,
		code2="is",
	),
	BabylonLanguage(
		name="Norwegian",
		charset="Latin",
		encoding="cp1252",
		code=0x2E,
		code2="no",
	),
	BabylonLanguage(
		name="Romanian",
		charset="Latin",
		encoding="cp1252",
		code=0x2F,
		code2="ro",
	),
	BabylonLanguage(
		name="Swedish",
		charset="Latin",
		encoding="cp1252",
		code=0x30,
		code2="sv",
	),
	BabylonLanguage(
		name="Ukrainian",
		charset="Cyrillic",
		encoding="cp1251",
		code=0x31,
		code2="uk",
	),
	BabylonLanguage(
		name="Belarusian",
		charset="Cyrillic",
		encoding="cp1251",
		code=0x32,
		code2="be",
	),
	BabylonLanguage(
		name="Persian",  # aka "Farsi"
		charset="Arabic",
		encoding="cp1256",
		code=0x33,
		code2="fa",
	),
	BabylonLanguage(
		name="Basque",
		charset="Latin",
		encoding="cp1252",
		code=0x34,
		code2="eu",
	),
	BabylonLanguage(
		name="Macedonian",
		charset="Eastern European",
		encoding="cp1250",
		code=0x35,
		code2="mk",
	),
	BabylonLanguage(
		name="Afrikaans",
		charset="Latin",
		encoding="cp1252",
		code=0x36,
		code2="af",
	),
	BabylonLanguage(
		# Babylon Glossary Builder spells this language "Faeroese"
		name="Faroese",
		charset="Latin",
		encoding="cp1252",
		code=0x37,
		code2="fo",
	),
	BabylonLanguage(
		name="Latin",
		charset="Latin",
		encoding="cp1252",
		code=0x38,
		code2="la",
	),
	BabylonLanguage(
		name="Esperanto",
		charset="Turkish",
		encoding="cp1254",
		code=0x39,
		code2="eo",
	),
	BabylonLanguage(
		name="Tamazight",
		# aka "Standard Moroccan Tamazight", "Standard Moroccan Berber"
		# or "Standard Moroccan Amazigh"
		charset="Latin",
		encoding="cp1252",
		code=0x3A,
		code2="zgh",  # no 2-letter code (ISO 639-1)
	),
	BabylonLanguage(
		name="Armenian",
		charset="Latin",
		encoding="cp1252",
		code=0x3B,
		code2="hy",
	),
	BabylonLanguage(
		name="Hindi",
		charset="Latin",
		encoding="cp1252",
		code=0x3C,
		code2="hi",
	),
	BabylonLanguage(
		name="Somali",
		charset="Latin",
		encoding="cp1252",
		code=0x3D,
		code2="so",
	),
)
languageByCode = {lang.code: lang for lang in languages}



# -*- coding: utf-8 -*-

import datetime as dt
import functools
import gzip
import io
import math
import pathlib
import struct
import typing
import zipfile
from typing import IO, TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
	from collections.abc import Callable
	from typing import Any, Literal

from pyglossary.core import log
from pyglossary.flags import NEVER
from pyglossary.glossary_types import EntryType, GlossaryType
from pyglossary.langs import langDict
from pyglossary.option import (
	Option,
	StrOption,
)
from pyglossary.plugin_lib import mutf8

__all__ = [
	"Reader",
	"Writer",
	"description",
	"enable",
	"extensionCreate",
	"extensions",
	"format",
	"kind",
	"lname",
	"optionsProp",
	"singleFile",
	"website",
	"wiki",
]

enable = True
lname = "quickdic6"
format = "QuickDic6"
description = "QuickDic version 6 (.quickdic)"
extensions = (".quickdic", ".quickdic.v006.zip")
extensionCreate = ".quickdic"
singleFile = True
sortOnWrite = NEVER

kind = "binary"
wiki = ""
website = (
	"https://github.com/rdoeffinger/Dictionary",
	"github.com/rdoeffinger/Dictionary",
)
# https://github.com/rdoeffinger/Dictionary/blob/master/dictionary-format-v6.txt
optionsProp: "dict[str, Option]" = {
	"normalizer_rules": StrOption(
		comment="ICU normalizer rules to use for index sorting",
	),
	"source_lang": StrOption(
		comment="The language of the tokens in the dictionary index",
	),
	"target_lang": StrOption(
		comment="The language of the dictionary entries",
	),
}

HASH_SET_INIT = (
	b"\xac\xed"  # magic
	b"\x00\x05"  # version
	b"\x73"  # object
	b"\x72"  # class
	# Java String "java.util.HashSet":
	b"\x00\x11\x6a\x61\x76\x61\x2e\x75\x74\x69"
	b"\x6c\x2e\x48\x61\x73\x68\x53\x65\x74"
)
"""First part of Java serialization of java.util.HashSet"""

HASH_SET_INIT2 = (
	# serialization ID:
	b"\xba\x44\x85\x95\x96\xb8\xb7\x34"
	b"\x03"  # flags: serialized, custom serialization function
	b"\x00\x00"  # fields count
	b"\x78"  # blockdata end
	b"\x70"  # null (superclass)
	b"\x77\x0c"  # blockdata short, 0xc bytes
)
"""Second part of Java serialization of java.util.HashSet"""

LINKED_HASH_SET_INIT = (
	(
		b"\xac\xed"  # magic
		b"\x00\x05"  # version
		b"\x73"  # object
		b"\x72"  # class
		# Java String "java.util.LinkedHashSet":
		b"\x00\x17\x6a\x61\x76\x61\x2e\x75\x74\x69"
		b"\x6c\x2e\x4c\x69\x6e\x6b\x65\x64"
		b"\x48\x61\x73\x68\x53\x65\x74"
		# serialization ID:
		b"\xd8\x6c\xd7\x5a\x95\xdd\x2a\x1e"
		b"\x02"  # flags
		b"\x00\x00"  # fields count
		b"\x78"  # blockdata end
		b"\x72"  # superclass (java.util.HashSet)
		b"\x00\x11\x6a\x61\x76\x61\x2e\x75\x74\x69"
		b"\x6c\x2e\x48\x61\x73\x68\x53\x65\x74"
	)
	+ HASH_SET_INIT2
)
"""Header of Java serialization of java.util.LinkedHashSet"""

HASH_SET_CAPACITY_FACTOR = 0.75
"""Capacity factor used to determine the hash set's capacity from its length"""

default_de_normalizer_rules = (
	":: Lower; 'ae' > 'ä'; 'oe' > 'ö'; 'ue' > 'ü'; 'ß' > 'ss'; "
)
default_normalizer_rules = (
	":: Any-Latin; ' ' > ; "
	":: Lower; :: NFD; "
	":: [:Nonspacing Mark:] Remove; "
	":: NFC ;"
)


def read_byte(fp: IO[bytes]) -> int:
	return struct.unpack(">b", fp.read(1))[0]


def write_byte(fp: IO[bytes], val: int) -> int:
	return fp.write(struct.pack(">b", val))


def read_bool(fp: IO[bytes]) -> bool:
	return bool(read_byte(fp))


def write_bool(fp: IO[bytes], val: int) -> int:
	return write_byte(fp, val)


def read_short(fp: IO[bytes]) -> int:
	return struct.unpack(">h", fp.read(2))[0]


def write_short(fp: IO[bytes], val: int) -> int:
	return fp.write(struct.pack(">h", val))


def read_int(fp: IO[bytes]) -> int:
	return struct.unpack(">i", fp.read(4))[0]


def write_int(fp: IO[bytes], val: int) -> int:
	return fp.write(struct.pack(">i", val))


def read_long(fp: IO[bytes]) -> int:
	return struct.unpack(">q", fp.read(8))[0]


def write_long(fp: IO[bytes], val: int) -> int:
	return fp.write(struct.pack(">q", val))


def read_float(fp: IO[bytes]) -> float:
	return struct.unpack(">f", fp.read(4))[0]


def write_float(fp: IO[bytes], val: float) -> int:
	return fp.write(struct.pack(">f", val))


def read_string(fp: IO[bytes]) -> str:
	length = read_short(fp)
	return mutf8.decode_modified_utf8(fp.read(length))


def write_string(fp: IO[bytes], val: str) -> int:
	b_string = mutf8.encode_modified_utf8(val)
	return write_short(fp, len(b_string)) + fp.write(b_string)


def read_hashset(fp: IO[bytes]) -> list[str]:
	hash_set_init = fp.read(len(HASH_SET_INIT))
	if hash_set_init == HASH_SET_INIT:
		hash_set_init2 = fp.read(len(HASH_SET_INIT2))
		assert hash_set_init2 == HASH_SET_INIT2
	else:
		n_extra = len(LINKED_HASH_SET_INIT) - len(HASH_SET_INIT)
		hash_set_init += fp.read(n_extra)
		assert hash_set_init == LINKED_HASH_SET_INIT
	read_int(fp)  # capacity
	capacity_factor = read_float(fp)
	assert capacity_factor == HASH_SET_CAPACITY_FACTOR
	num_entries = read_int(fp)
	data: list[str] = []
	while len(data) < num_entries:
		assert read_byte(fp) == 0x74
		data.append(read_string(fp))
	assert read_byte(fp) == 0x78
	return data


def write_hashset(
	fp: IO[bytes],
	data: list[str],
	linked_hash_set: bool = False,
) -> int:
	write_start_offset = fp.tell()
	if linked_hash_set:
		fp.write(LINKED_HASH_SET_INIT)
	else:
		fp.write(HASH_SET_INIT + HASH_SET_INIT2)
	num_entries = len(data)
	capacity = (
		2 ** math.ceil(math.log2(num_entries / HASH_SET_CAPACITY_FACTOR))
		if num_entries > 0
		else 128
	)
	write_int(fp, capacity)
	write_float(fp, HASH_SET_CAPACITY_FACTOR)
	write_int(fp, num_entries)
	for string in data:
		write_byte(fp, 0x74)
		write_string(fp, string)
	write_byte(fp, 0x78)
	return fp.tell() - write_start_offset


T = TypeVar("T")


def read_list(
	fp: IO[bytes],
	fun: "Callable[[IO[bytes]], T]",
) -> list[T]:
	size = read_int(fp)
	toc = struct.unpack(f">{size + 1}q", fp.read(8 * (size + 1)))
	entries = []
	for offset in toc[:-1]:
		fp.seek(offset)
		entries.append(fun(fp))
	fp.seek(toc[-1])
	return entries


def write_list(
	fp: IO[bytes],
	fun: "Callable[[IO[bytes], T], Any]",
	entries: list[T],
) -> int:
	write_start_offset = fp.tell()
	size = len(entries)
	write_int(fp, size)
	toc_offset = fp.tell()
	fp.seek(toc_offset + 8 * (size + 1))
	toc = [fp.tell()]
	for e in entries:
		fun(fp, e)
		toc.append(fp.tell())
	fp.seek(toc_offset)
	fp.write(struct.pack(f">{size + 1}q", *toc))
	fp.seek(toc[-1])
	return fp.tell() - write_start_offset


def read_entry_int(fp: IO[bytes]) -> int:
	return read_int(fp)


def write_entry_int(fp: IO[bytes], entry: int) -> int:
	return write_int(fp, entry)


def read_entry_source(fp: IO[bytes]) -> tuple[str, int]:
	name = read_string(fp)
	count = read_int(fp)
	return name, count


def write_entry_source(fp: IO[bytes], entry: tuple[str, int]) -> int:
	name, count = entry
	return write_string(fp, name) + write_int(fp, count)


def read_entry_pairs(fp: IO[bytes]) -> tuple[int, list[tuple[str, str]]]:
	src_idx = read_short(fp)
	count = read_int(fp)
	pairs = [(read_string(fp), read_string(fp)) for i in range(count)]
	return src_idx, pairs


def write_entry_pairs(
	fp: IO[bytes],
	entry: tuple[int, list[tuple[str, str]]],
) -> int:
	write_start_offset = fp.tell()
	src_idx, pairs = entry
	write_short(fp, src_idx)
	write_int(fp, len(pairs))
	for p in pairs:
		write_string(fp, p[0])
		write_string(fp, p[1])
	return fp.tell() - write_start_offset


def read_entry_text(fp: IO[bytes]) -> tuple[int, str]:
	src_idx = read_short(fp)
	txt = read_string(fp)
	return src_idx, txt


def write_entry_text(fp: IO[bytes], entry: tuple[int, str]) -> int:
	src_idx, txt = entry
	return write_short(fp, src_idx) + write_string(fp, txt)


def read_entry_html(fp: IO[bytes]) -> tuple[int, str, str]:
	src_idx = read_short(fp)
	title = read_string(fp)
	read_int(fp)  # len_raw
	len_compr = read_int(fp)
	b_compr = fp.read(len_compr)
	with gzip.open(io.BytesIO(b_compr), "r") as zf:
		# this is not modified UTF-8 (read_string), but actual UTF-8
		html = zf.read().decode()
	return src_idx, title, html


def write_entry_html(fp: IO[bytes], entry: tuple[int, str, str]) -> int:
	write_start_offset = fp.tell()
	src_idx, title, html = entry
	b_html = "".join(c if ord(c) < 128 else f"&#{ord(c)};" for c in html).encode()
	ib_compr = io.BytesIO()
	with gzip.GzipFile(fileobj=ib_compr, mode="wb") as zf:
		# note that the compressed bytes might differ from the original Java
		# implementation that uses GZIPOutputStream
		zf.write(b_html)
	ib_compr.seek(0)
	b_compr = ib_compr.read()
	write_short(fp, src_idx)
	write_string(fp, title)
	write_int(fp, len(b_html))
	write_int(fp, len(b_compr))
	fp.write(b_compr)
	return fp.tell() - write_start_offset


IndexEntryType = tuple[
	str,  # token
	int,  # start_index
	int,  # count
	str,  # token_norm
	list[int],  # html_indices
]


EntryIndexTuple = tuple[
	str,  # short_name
	str,  # long_name
	str,  # iso
	str,  # normalizer_rules
	bool,  # swap_flag
	int,  # main_token_count
	list[IndexEntryType],  # index_entries
	list[str],  # stop_list,
	list[tuple[int, int]],  # rows
]


def read_entry_index(fp: IO[bytes]) -> EntryIndexTuple:
	short_name = read_string(fp)
	long_name = read_string(fp)
	iso = read_string(fp)
	normalizer_rules = read_string(fp)
	swap_flag = read_bool(fp)
	main_token_count = read_int(fp)
	index_entries = read_list(fp, read_entry_indexentry)

	stop_list_size = read_int(fp)
	stop_list_offset = fp.tell()
	stop_list = read_hashset(fp)
	assert fp.tell() == stop_list_offset + stop_list_size

	num_rows = read_int(fp)
	row_size = read_int(fp)
	row_data = fp.read(num_rows * row_size)
	rows = [
		# <type>, <index>
		struct.unpack(">bi", row_data[j : j + row_size])
		for j in range(0, len(row_data), row_size)
	]
	return (
		short_name,
		long_name,
		iso,
		normalizer_rules,
		swap_flag,
		main_token_count,
		index_entries,
		stop_list,
		rows,
	)


def write_entry_index(
	fp: IO[bytes],
	entry: EntryIndexTuple,
) -> int:
	write_start_offset = fp.tell()
	(
		short_name,
		long_name,
		iso,
		normalizer_rules,
		swap_flag,
		main_token_count,
		index_entries,
		stop_list,
		rows,
	) = entry
	write_string(fp, short_name)
	write_string(fp, long_name)
	write_string(fp, iso)
	write_string(fp, normalizer_rules)
	write_bool(fp, swap_flag)
	write_int(fp, main_token_count)
	write_list(fp, write_entry_indexentry, index_entries)

	stop_list_size_offset = fp.tell()
	stop_list_offset = stop_list_size_offset + write_int(fp, 0)
	stop_list_size = write_hashset(fp, stop_list, linked_hash_set=True)
	fp.seek(stop_list_size_offset)
	write_int(fp, stop_list_size)
	fp.seek(stop_list_offset + stop_list_size)

	write_int(fp, len(rows))
	write_int(fp, 5)
	row_data = b"".join([struct.pack(">bi", t, i) for t, i in rows])
	fp.write(row_data)
	return fp.tell() - write_start_offset


def read_entry_indexentry(fp: IO[bytes]) -> IndexEntryType:
	token = read_string(fp)
	start_index = read_int(fp)
	count = read_int(fp)
	has_normalized = read_bool(fp)
	token_norm = read_string(fp) if has_normalized else ""
	html_indices = read_list(fp, read_entry_int)
	return token, start_index, count, token_norm, html_indices


def write_entry_indexentry(
	fp: IO[bytes],
	entry: IndexEntryType,
) -> None:
	token, start_index, count, token_norm, html_indices = entry
	has_normalized = bool(token_norm)
	write_string(fp, token)
	write_int(fp, start_index)
	write_int(fp, count)
	write_bool(fp, has_normalized)
	if has_normalized:
		write_string(fp, token_norm)
	write_list(fp, write_entry_int, html_indices)


class Comparator:
	def __init__(self, locale_str: str, normalizer_rules: str, version: int) -> None:
		import icu

		self.version = version
		self.locale = icu.Locale(locale_str)
		self._comparator = (
			icu.RuleBasedCollator("&z<ȝ")
			if self.locale.getLanguage() == "en"
			else icu.Collator.createInstance(self.locale)
		)
		self._comparator.setStrength(icu.Collator.IDENTICAL)
		self.normalizer_rules = normalizer_rules
		self.normalize = icu.Transliterator.createFromRules(
			"",
			self.normalizer_rules,
			icu.UTransDirection.FORWARD,
		).transliterate

	def compare(
		self,
		tup1: "tuple[str, str]",
		tup2: "tuple[str, str]",
	) -> "Literal[0, 1, -1]":
		# assert isinstance(tup1, tuple)
		# assert isinstance(tup2, tuple)
		s1, n1 = tup1
		s2, n2 = tup2
		cn = self._compare_without_dash(n1, n2)
		if cn != 0:
			return cn
		cn = self._comparator.compare(n1, n2)
		if cn != 0 or self.version < 7:
			return cn
		return self._comparator.compare(s1, s2)

	def _compare_without_dash(
		self,
		a: str,
		b: str,
	) -> "Literal[0, 1, -1]":
		if self.version < 7:
			return 0
		s1 = self._without_dash(a)
		s2 = self._without_dash(b)
		return self._comparator.compare(s1, s2)

	@staticmethod
	def _without_dash(a: str) -> str:
		return a.replace("-", "").replace("þ", "th").replace("Þ", "Th")


class QuickDic:
	def __init__(  # noqa: PLR0913
		self,
		name: str,
		sources: list[tuple[str, int]],
		pairs: list[tuple[int, list[tuple[str, str]]]],
		texts: list[tuple[int, str]],
		htmls: list[tuple[int, str, str]],
		version: int = 6,
		indices: list[EntryIndexTuple] | None = None,
		created: dt.datetime | None = None,
	) -> None:
		self.name = name
		self.sources = sources
		self.pairs = pairs
		self.texts = texts
		self.htmls = htmls
		self.version = version
		self.indices = [] if indices is None else indices
		self.created = dt.datetime.now() if created is None else created

	@classmethod
	def from_path(cls: "type[QuickDic]", path_str: str) -> "QuickDic":
		path = pathlib.Path(path_str)
		if path.suffix != ".zip":
			with open(path, "rb") as fp:
				return cls.from_fp(fp)
		with zipfile.ZipFile(path, mode="r") as zf:
			fname = next(n for n in zf.namelist() if n.endswith(".quickdic"))
			with zf.open(fname) as fp:
				return cls.from_fp(fp)

	@classmethod
	def from_fp(cls: "type[QuickDic]", fp: IO[bytes]) -> "QuickDic":
		version = read_int(fp)
		created = dt.datetime.fromtimestamp(float(read_long(fp)) / 1000.0)  # noqa: DTZ006
		name = read_string(fp)
		sources = read_list(fp, read_entry_source)
		pairs = read_list(fp, read_entry_pairs)
		texts = read_list(fp, read_entry_text)
		htmls = read_list(fp, read_entry_html)
		indices = read_list(fp, read_entry_index)
		assert read_string(fp) == "END OF DICTIONARY"
		return cls(
			name,
			sources,
			pairs,
			texts,
			htmls,
			version=version,
			indices=indices,
			created=created,
		)

	def add_index(  # noqa: PLR0913
		self,
		short_name: str,
		long_name: str,
		iso: str,
		normalizer_rules: str,
		synonyms: dict | None = None,
	) -> None:
		swap_flag = False
		comparator = Comparator(iso, normalizer_rules, self.version)

		synonyms = synonyms or {}
		n_synonyms = sum(len(v) for v in synonyms.values())
		log.info(f"Adding an index for {iso} with {n_synonyms} synonyms ...")

		# since we don't tokenize, the stop list is always empty
		stop_list: list[str] = []
		if self.indices is None:
			self.indices = []

		log.info("Initialize token list ...")
		tokens1 = [
			(pair[1 if swap_flag else 0], 0, idx)
			for idx, (_, pairs) in enumerate(self.pairs)
			for pair in pairs
		]
		if not swap_flag:
			tokens1.extend(
				[(title, 4, idx) for idx, (_, title, _) in enumerate(self.htmls)],
			)
		tokens1 = [(t.strip(), ttype, tidx) for t, ttype, tidx in tokens1]

		log.info("Normalize tokens ...")
		tokens = [
			(t, comparator.normalize(t), ttype, tidx) for t, ttype, tidx in tokens1 if t
		]

		if len(synonyms) > 0:
			log.info(
				f"Insert synonyms into token list ({len(tokens)} entries) ...",
			)
			tokens.extend(
				[
					(s, comparator.normalize(s)) + t[2:]
					for t in tokens
					if t[0] in synonyms
					for s in synonyms[t[0]]
					if s
				],
			)

		log.info(f"Sort tokens with synonyms ({len(tokens)} entries) ...")
		key_fun = functools.cmp_to_key(comparator.compare)
		tokens.sort(key=lambda t: key_fun((t[0], t[1])))

		log.info("Build mid-layer index ...")
		rows: list[tuple[int, int]] = []
		index_entries: list[IndexEntryType] = []
		for token, token_norm, ttype, tidx in tokens:
			prev_token = "" if len(index_entries) == 0 else index_entries[-1][0]
			if prev_token == token:
				(
					token,  # noqa: PLW2901
					index_start,
					count,
					token_norm,  # noqa: PLW2901
					html_indices,
				) = index_entries.pop()
			else:
				i_entry = len(index_entries)
				index_start = len(rows)
				count = 0
				token_norm = "" if token == token_norm else token_norm  # noqa: PLW2901
				html_indices = []
				rows.append((1, i_entry))
			if ttype == 4:
				if tidx not in html_indices:
					html_indices.append(tidx)
			elif (ttype, tidx) not in rows[index_start + 1 :]:
				rows.append((ttype, tidx))
				count += 1
			index_entries.append(
				(token, index_start, count, token_norm, html_indices),
			)

		# the exact meaning of this parameter is unknown,
		# and it seems to be ignored by readers
		main_token_count = len(index_entries)

		self.indices.append(
			(
				short_name,
				long_name,
				iso,
				normalizer_rules,
				swap_flag,
				main_token_count,
				index_entries,
				stop_list,
				rows,
			),
		)

	def write(self, path: str) -> None:
		with open(path, "wb") as fp:
			log.info(f"Writing to {path} ...")
			write_int(fp, self.version)
			write_long(fp, int(self.created.timestamp() * 1000))
			write_string(fp, self.name)
			write_list(fp, write_entry_source, self.sources)
			write_list(fp, write_entry_pairs, self.pairs)
			write_list(fp, write_entry_text, self.texts)
			write_list(fp, write_entry_html, self.htmls)
			write_list(fp, write_entry_index, self.indices)
			write_string(fp, "END OF DICTIONARY")


class Reader:
	depends = {
		"icu": "PyICU",
	}

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._dic: QuickDic | None = None

	def open(self, filename: str) -> None:
		self._filename = filename
		self._dic = QuickDic.from_path(self._filename)
		self._glos.setDefaultDefiFormat("h")
		self._extract_synonyms_from_indices()

	def _extract_synonyms_from_indices(self) -> None:
		self._text_tokens: dict[int, str] = {}
		self._synonyms: dict[tuple[int, int], set[str]] = {}
		assert self._dic is not None
		for index in self._dic.indices:
			_, _, _, _, swap_flag, _, index_entries, _, _ = index

			# Note that we ignore swapped indices because pyglossary assumes
			# uni-directional dictionaries.
			# It might make sense to add an option in the future to read only the
			# swapped indices (create a dictionary with reversed direction).
			if swap_flag:
				continue

			for i_entry, index_entry in enumerate(index_entries):
				e_rows = self._extract_rows_from_indexentry(index, i_entry)
				token, _, _, token_norm, _ = index_entry
				for entry_id in e_rows:
					if entry_id not in self._synonyms:
						self._synonyms[entry_id] = set()
					self._synonyms[entry_id].add(token)
					if token_norm:
						self._synonyms[entry_id].add(token_norm)

	def _extract_rows_from_indexentry(
		self,
		index: EntryIndexTuple,
		i_entry: int,
		recurse: list[int] | None = None,
	) -> list[tuple[int, int]]:
		recurse = recurse or []
		recurse.append(i_entry)
		_, _, _, _, _, _, index_entries, _, rows = index
		token, start_index, count, _, html_indices = index_entries[i_entry]
		block_rows = rows[start_index : start_index + count + 1]
		assert block_rows[0][0] in {1, 3}
		assert block_rows[0][1] == i_entry
		e_rows = []
		for entry_type, entry_idx in block_rows[1:]:
			if entry_type in {1, 3}:
				# avoid an endless recursion
				if entry_idx not in recurse:
					e_rows.extend(
						self._extract_rows_from_indexentry(
							index,
							entry_idx,
							recurse=recurse,
						),
					)
			else:
				e_rows.append((entry_type, entry_idx))
				if entry_type == 2 and entry_idx not in self._text_tokens:
					self._text_tokens[entry_idx] = token
		for idx in html_indices:
			if (4, idx) not in e_rows:
				e_rows.append((4, idx))
		return e_rows

	def close(self) -> None:
		self.clear()

	def clear(self) -> None:
		self._filename = ""
		self._dic = None

	def __len__(self) -> int:
		if self._dic is None:
			return 0
		return sum(len(p) for _, p in self._dic.pairs) + len(self._dic.htmls)

	def __iter__(self) -> "typing.Iterator[EntryType]":
		if self._dic is None:
			raise RuntimeError("dictionary not open")
		for idx, (_, pairs) in enumerate(self._dic.pairs):
			syns = self._synonyms.get((0, idx), set())
			for word, defi in pairs:
				l_word = [word] + sorted(syns.difference({word}))
				yield self._glos.newEntry(l_word, defi, defiFormat="m")
		for idx, (_, defi) in enumerate(self._dic.texts):
			if idx not in self._text_tokens:
				# Ignore this text entry since it is not mentioned in the index at all
				# so that we don't even have a token or title for it.
				continue
			word = self._text_tokens[idx]
			syns = self._synonyms.get((2, idx), set())
			l_word = [word] + sorted(syns.difference({word}))
			yield self._glos.newEntry(l_word, defi, defiFormat="m")
		for idx, (_, word, defi) in enumerate(self._dic.htmls):
			syns = self._synonyms.get((4, idx), set())
			l_word = [word] + sorted(syns.difference({word}))
			yield self._glos.newEntry(l_word, defi, defiFormat="h")


class Writer:
	_normalizer_rules = ""
	_source_lang = ""
	_target_lang = ""

	def __init__(self, glos: GlossaryType) -> None:
		self._glos = glos
		self._filename = ""
		self._dic = None

	def finish(self) -> None:
		self._filename = ""
		self._dic = None

	def open(self, filename: str) -> None:
		self._filename = filename

	def write(self) -> "typing.Generator[None, EntryType, None]":
		synonyms: dict[str, list[str]] = {}
		htmls = []
		log.info("Converting individual entries ...")
		while True:
			entry = yield
			if entry is None:
				break
			if entry.isData():
				log.warn(f"Ignoring binary data entry {entry.l_word[0]}")
				continue

			entry.detectDefiFormat()
			if entry.defiFormat not in {"h", "m"}:
				log.error(f"Unsupported defiFormat={entry.defiFormat}, assuming 'h'")

			words = entry.l_word
			if words[0] in synonyms:
				synonyms[words[0]].extend(words[1:])
			else:
				synonyms[words[0]] = words[1:]

			# Note that we currently write out all entries as "html" type entries.
			# In the future, it might make sense to add an option that somehow
			# specifies the entry type to use.
			htmls.append((0, words[0], entry.defi))

		log.info("Collecting meta data ...")
		name = self._glos.getInfo("bookname")
		if not name:
			name = self._glos.getInfo("description")

		sourceLang = (
			self._glos.sourceLang
			if not self._source_lang
			else langDict[self._source_lang]
		)
		targetLang = (
			self._glos.targetLang
			if not self._target_lang
			else langDict[self._target_lang]
		)
		if sourceLang and targetLang:
			sourceLang = sourceLang.code
			targetLang = targetLang.code
		else:
			# fallback if no languages are specified
			sourceLang = targetLang = "EN"
		langs = f"{sourceLang}->{targetLang}"
		if langs not in name.lower():
			name = f"{self._glos.getInfo('name')} ({langs})"

		sources = [("", len(htmls))]
		pairs = []
		texts = []
		self._dic = QuickDic(name, sources, pairs, texts, htmls)

		short_name = long_name = iso = sourceLang
		normalizer_rules = (
			self._normalizer_rules or default_de_normalizer_rules
			if iso == "DE"
			else default_normalizer_rules
		)
		self._dic.add_index(
			short_name,
			long_name,
			iso,
			normalizer_rules,
			synonyms=synonyms,
		)

		self._dic.write(self._filename)



from typing import TYPE_CHECKING

from pyglossary.sort_modules import ebook

if TYPE_CHECKING:
	from .sort_keys_types import sortKeyType, sqliteSortKeyType


desc = "E-Book (prefix length: 3)"


def normal(sortEncoding: str = "utf-8", **_options) -> "sortKeyType":
	return ebook.normal(
		sortEncoding=sortEncoding,
		group_by_prefix_length=3,
	)


def sqlite(
	sortEncoding: str = "utf-8",
	**_options,
) -> "sqliteSortKeyType":
	return ebook.sqlite(
		sortEncoding=sortEncoding,
		group_by_prefix_length=3,
	)
